/**
  Tiny OpenGL/GLUT particle system example program
  
  This version:
  	- Uses second-order dynamics
	- Treats "Z" as up (the right choice!)
	- Allows nice camera rotation using three vectors
	- Allows camera motion with "WASD" keys
	- Draws particles into a little ground texture, updated with a framebuffer object
  
  Dr. Orion Sky Lawlor, olawlor@acm.org, 2009-01-21 (Public Domain)
*/
#include "GL/glew.h" /* GL extension wrangler: http://glew.sourceforge.net/ */
#include "GL/glut.h" /* windows, mice, etc.  http://freeglut.sourceforge.net */
#include <stdlib.h> /* for "rand" */
#include <stdio.h> /* for printf */
#include "osl/vec4.h" /* 3D vector class */
#include "ogl/glsl.h" /* programmable shaders */
#include "ogl/glsl.cpp"

#define VEC3_TO_XYZ(v) (v).x,(v).y,(v).z /* convert a vec3 to three float arguments */
#include <vector> /* std::vector container */

/* Random Float: Return a random number between 0 and 1 */
float rf(void) { return (rand()&0xfff)*(1.0/0xfff); }
/* Random Vector: return a random vector in 3D, all axes between 0 and 1 */
vec3 rv(void) {return vec3(rf(),rf(),rf());}

bool key_down[256]; /* if true, the corresponding key is down */

float rotate_x=0, rotate_y=0;

/* Return the wall clock time, in seconds */
double oglTimer(void) {return 0.001*glutGet(GLUT_ELAPSED_TIME);}
/* Check if an error has happened--if so, exit */
void oglCheck(const char *where) {
	int err=0;
	if (0!=(err=glGetError())) {
		fprintf(stderr,"OpenGL error %d at %s\n",err,where);
		exit(1);
	}
}

class particle {
public:
	vec3 P; /* position */
	vec3 V; /* velocity */
	vec3 C; /* color */
};


/* 3D orientation */
/* Orthonormal coordinate frame */
class ortho_frame {
public:
	/* Unit vectors pointing along axes of our frame.
		X cross Y is Z. (right-handed coordinate system)
	*/
	vec3 x,y,z;
	ortho_frame() :x(1,0,0), y(0,0,1), z(0,-1,0) {} /* point down Y axis */
	
	/* Reorient this coordinate frame by this far along in the X and Y axes.  
	   "dx" is the distance along the z axis to push the x axis; 
	   "dy" the distance along the z axis to push the y axis.
	*/
	void nudge(double dx,double dy) {
		x+=dx*z;
		y+=dy*z;
		orthonormalize();
	}
	
	/* Reconstruct an orthonormal frame from X and Y axes.
	  Y is primary, X is secondary, Z is tertiary.
	*/
	void orthonormalize(void) {
		y=normalize(y);
		z=normalize(cross(x,y));
		x=normalize(cross(y,z));
	}
};
ortho_frame orient; /* camera orientation */
vec3 camera(0,-1.0,0.5); /* (initial) location of camera */


/* Return the distance between these two pointers, in bytes,
itself expressed as a pointer.  Needed for glVertexPointer &c. */
void *ptr_diff(void *a,void *b) {
	return (void *)(long)((char *)a - (char *)b);
}

/* Make an OpenGL texture handle, of this size, with this pixel format */
GLuint make_texture(int w,int h,GLenum format=GL_RGBA8)
{
	GLuint tex=0;
	glGenTextures(1,&tex);
	GLenum t=GL_TEXTURE_2D;
	glBindTexture(t,tex);
	char *mem=new char[w*h*4]; memset(mem,0x80,w*h*4);
	glTexImage2D(t,0,format, w,h,0, GL_RGBA,GL_UNSIGNED_BYTE,mem);
	delete[] mem;
	glTexParameteri(t, GL_TEXTURE_MAG_FILTER,GL_LINEAR); 
	glTexParameteri(t, GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR); 
	
	glTexParameterf(t, GL_TEXTURE_MIN_LOD, 0.0);
	glGenerateMipmapEXT(t);
	
	GLenum texWrap=GL_CLAMP_TO_EDGE; /* stop changing at edge */
	glTexParameteri(t, GL_TEXTURE_WRAP_S, texWrap); 
	glTexParameteri(t, GL_TEXTURE_WRAP_T, texWrap);
	glBindTexture(t,0); /* back to default texture */
	oglCheck("after make_texture");
	return tex;
}

/* Make a framebuffer object that renders pixels into this texture */
GLuint make_fbo(GLuint tex)
{
	GLuint fb=0;
	glGenFramebuffersEXT(1, &fb);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fb);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, 
		GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, tex, 0);
	
	/* Make sure the framebuffer's happy */
	GLenum status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
	if (status!=GL_FRAMEBUFFER_COMPLETE_EXT)
	{ fprintf(stderr,"Error %d in framebuffer creation!\n",status); exit(1); }

	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0); /* back to screen render */
	oglCheck("after make_fbo");
	return fb;
}

/* Erase this framebuffer object to this color */
void erase_fbo(GLuint fb,vec4 color) {
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fb);
	glClearColor(color.x,color.y,color.z,color.w);
	glClear(GL_COLOR_BUFFER_BIT);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0); /* back to screen render */
	oglCheck("after erase_fbo");
}

void display(void) /* Draw everything in our window */
{
	oglCheck("before display function");
	static double elapsed=0.0; /* last elapsed display time (seconds) */
	static float dt=0.01, display_dt=0.0;
	static float sim_time=0.0;
	float big=30; /* size of the ground */
	float floor=0.0; /* Z height of floor */
	glFinish(); double start_t=oglTimer();
	
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	
/* Particle state storage */
	static bool first_time=true; /* for initial bc's! */
	static std::vector<particle> pts;
	if (first_time) { /* First time: initialize points */
		static int count=0;
		srand(count++);
		for (int i=0;i<100000;i++) {
			particle p;
			p.P=rv(); 
			p.P.z=0.5*rf()+0.1; /* squish & shift Z */
			//p.P.x=p.P.z; /* z == initial Y */
			p.V=0.3*(rv()-vec3(0.5)); 
			p.V=vec3(0.0); /* don't move initially */
			p.C=rv();
			pts.push_back(p);
		}
		first_time=true;
	}
	
	static float energy=0, m=1.0; /* mass and energy (for energy accounting) */
	
	static int dots_pix=1024; /* pixels in our little dots texture */
	static GLuint dots_tex=make_texture(dots_pix,dots_pix,GL_RGBA8);
	static GLuint dots_fbo=make_fbo(dots_tex);
	static int fbo_setup=1;
	if (fbo_setup) erase_fbo(dots_fbo,vec4(1,0,0,1)); /* opaque red */
	fbo_setup=0;
	
	/* Bind the dots texture framebuffer, so we can draw our little dots drops */
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, dots_fbo);
	glViewport(0,0,dots_pix,dots_pix);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	/* From 0..1 texture coords to -1..+1 clip coords */
	glTranslatef(-1.0,-1.0,-1.0);
	glScalef(2.0,2.0,2.0);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glEnable(GL_BLEND);
	glEnable(GL_POINT_SMOOTH);
	glPointSize(5.0);
	glBegin(GL_POINTS);
	glColor4f(0,0,0,1.0); /* black splatty points */
	/* Subcycle, so the physics runs at the right stepsize, 
	   yet still keeps up with the display. */
	double wall_time=oglTimer(); // *0.1; // <- for slow-motion 
	while (sim_time<wall_time && !key_down[(int)' '])
	{ /* take one physics step */
		sim_time+=dt;
		energy=0;
		for (int i=0;i<pts.size();i++) {
			particle &p=pts[i];

		/* Apply physics */
			vec3 gA(0.0,0.0,-1.0); /* gravity acceleration */

			float KE=0.5*m*dot(p.V,p.V);
			float PE=m*dot(gA,-p.P);
			energy+=PE+KE;
			
			vec3 fA=-0.01*p.V; /* frictional acceleration (a=F/m) */
			vec3 A=gA+fA; /* total acceleration */

			p.P+=dt*p.V+0.5*dt*dt*A; /* move position by velocity */
			p.V+=dt*A; /* accelerate particles downwards */

			float wall=0.0;
			if (p.P.x<wall) {
				p.V.x=+fabs(p.V.x);
			}
			wall=+1.0;
			if (p.P.x>wall) {
				p.V.x=-fabs(p.V.x);
			}
			
			float r=0.2;
			vec3 center(0.5,0.5,0.05);
			vec3 d=p.P-center;
			float pr=length(d);
			if (pr<r) { /* our point is inside the sphere--bounce off */
				/* Reflect velocity around d vector: */
				vec3 nd=d*(1.0/pr); /* normalized d vector == sphere normal (facing outwards) */
				p.V=p.V-2.0*nd*dot(p.V,nd);
				if (first_time) p.P=center+d*(r/pr); /* scale to surface */
			}
			
			if (p.P.z<floor) 
			{ /* "splat" on the floor */
				//printf("Splatting at %f, %f\n",p.P.x,p.P.y);
				//glColor3fv(p.C);
				glVertex3f(p.P.x,p.P.y,0);
				int last=pts.size()-1; /* swap out us with last point */
				pts[i]=pts[last];
				pts.erase(pts.begin()+last);
			}
		}

		first_time=false;
	}
	glEnd();
	

/* Main camera setup */
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
	glViewport(0,0,glutGet(GLUT_WINDOW_WIDTH),glutGet(GLUT_WINDOW_HEIGHT));
	glClearColor(0.3,0.5,0.7,1.0); /* clear to a nice blue color */
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(70.0, /* degree FOV */
		glutGet(GLUT_WINDOW_WIDTH)/float(glutGet(GLUT_WINDOW_HEIGHT)),
		0.001,1000.0 /* Z range */
	);
	
	/* Physics and display timesteps: */
	float vel=3.0*display_dt; /* meters camera motion per frame */
	if (key_down[(int)'w']) camera-=vel*orient.z;
	if (key_down[(int)'s']) camera+=vel*orient.z;
	if (key_down[(int)'a']) camera-=vel*orient.x;
	if (key_down[(int)'d']) camera+=vel*orient.x;
	
	/* Don't allow head tilting! */
	orient.x.z=0.0; 
	orient.y=cross(orient.z,orient.x);
	orient.orthonormalize();
	
	vec3 target=camera-orient.z; /* camera looks down *negative* Z */
	gluLookAt(VEC3_TO_XYZ(camera), 
		VEC3_TO_XYZ(target), 
		VEC3_TO_XYZ(orient.y)
	);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	// Draw in a ground plane
	glEnable(GL_DEPTH_TEST); /* read/write from depth buffer */
	glDepthMask(GL_FALSE); /* don't write to depth buffer (no Z fighting!) */
	
	glBegin(GL_QUADS);
	glColor4f(0.6,0.4,0.2,1.0); /* light, opaque brown */
	glVertex3f(-big,-big,floor);
	glVertex3f(+big,-big,floor);
	glVertex3f(+big,+big,floor);
	glVertex3f(-big,+big,floor);
	glEnd();
	
	// Draw our dots texture on the ground, with blurring
	
	static GLhandleARB prog=makeProgramObject(
		"varying vec2 tc; /* texture coordinates */\n"
		"void main(void) { /* vertex shader */ \n"
		"	tc = vec2(gl_MultiTexCoord0);\n"
		"	gl_Position = gl_ModelViewProjectionMatrix * gl_Vertex;\n"
		"}\n"
		,
		"varying vec2 tc; /* texture coordinates */\n"
		"uniform sampler2D myTex;\n"
		"void main(void) { /* fragment shader */\n"
		"	vec4 sum=vec4(0.0);\n"
		"	float totalWeight=0.0;\n"
		"	int sz=1; /* pixels on each side for kernel */\n"
		"	for (int y=-sz;y<=+sz;y++)\n"
		"	for (int x=-sz;x<=+sz;x++) {\n"
		"		vec2 shift=vec2(x,y)*(100.0/1024);\n"
		"		float w=1.0;\n"
		"		sum+=w*texture2D(myTex,tc+shift);\n"
		"		totalWeight+=w;\n"
		"	}\n"
		"	\n"
		"	gl_FragColor = sum/totalWeight;\n"
		"}\n"
	);
	glUseProgramObjectARB(prog);
	glActiveTextureARB(GL_TEXTURE0); /* put our texture into unit 0 */
	glBindTexture(GL_TEXTURE_2D,dots_tex);
	glGenerateMipmapEXT(GL_TEXTURE_2D);
	glUniform1iARB(glGetUniformLocationARB(prog,"myTex"),0); /* texture unit */
	
	glEnable(GL_BLEND);
	float sz=1.0;
	glBegin(GL_QUADS);
	glTexCoord2fv(vec3(0,0,0)); glVertex3f(0,0,floor);
	glTexCoord2fv(vec3(1,0,0)); glVertex3f(sz,0,floor);
	glTexCoord2fv(vec3(1,1,0)); glVertex3f(sz,sz,floor);
	glTexCoord2fv(vec3(0,1,0)); glVertex3f(0,sz,floor);
	glEnd();
	
	glUseProgramObjectARB(0);
	glDisable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);

	// Draw in a 1-meter grid on the ground
	glEnable(GL_LINE_SMOOTH);
	glBegin(GL_LINES);
	glColor4f(0.0,0.0,0.0,0.1); /* transparent black -> grey */
	for (int v=-big;v<=big;v++) {
		glVertex3f(-big,v,floor); /* X lines */
		glVertex3f(+big,v,floor);
		glVertex3f(v,-big,floor); /* Y lines */
		glVertex3f(v,+big,floor);
	}
	glEnd();
	
	glDepthMask(GL_TRUE); /* depth writes back on */
	
/* Draw axes */
	for (int axis=0;axis<3;axis++) {
		vec3 color(0.0);
		color[axis]=1.0;
		glColor3fv(color);
		
		glBegin(GL_LINES);
		glVertex3f(0,0,0);
		vec3 end(0.0); end[axis]=1.0;
		glVertex3fv(end);
		glEnd();
		
		glRasterPos3fv(end);
		glutBitmapCharacter(GLUT_BITMAP_TIMES_ROMAN_24,'X'+axis);
	}
	
/* Draw our points */
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_POINT_SMOOTH);
	glPointSize(8.4);
	/* Make gl_PointSize work from GLSL */
        glEnable(GL_VERTEX_PROGRAM_POINT_SIZE);
	
	static GLhandleARB simplePoint=makeProgramObject(
	"varying vec4 drawColor;\n"
	"uniform float pixPerParticle;\n"
	"void main(void) {\n"
	"	drawColor = gl_Color;\n"
	"	gl_Position = gl_ModelViewProjectionMatrix * gl_Vertex;\n"
	"	gl_PointSize = pixPerParticle/gl_Position.w;\n"
	"}\n"
	,
	"varying vec4 drawColor;\n"
	"void main(void) {\n"
	"	gl_FragColor = drawColor; \n"
	"}\n"
	);
	glUseProgramObjectARB(simplePoint);
	glUniform1fARB(glGetUniformLocationARB(simplePoint,"pixPerParticle"),
		6.0e-3*glutGet(GLUT_WINDOW_HEIGHT));
	
	glBegin(GL_POINTS);
	for (int i=0;i<pts.size();i++) {
		particle &p=pts[i];
	/* Draw the particle */
		glColor3fv(p.C);
		glVertex3fv(p.P);
	}
	glEnd();
	glUseProgramObjectARB(0); /* back to plain non-programmable OpenGL */
	
	glutSwapBuffers(); /* display everything we've rendered so far */
	glutPostRedisplay(); /* continual animation */
	
	// Print frames per second
	glFinish(); elapsed=oglTimer()-start_t; /* stop the stopwatch */
	static double last_print=0.0; /* Periodically print out the time */
	if (start_t>last_print+0.1) {
		display_dt=elapsed;
		printf("energy = %g\n",energy);
		printf("%.1f frames/sec, %.2f ms/frame\n",
			1.0/elapsed,1000.0*elapsed);
		last_print=start_t;
	}
	oglCheck("after display function");
}

int mouse_x=0,mouse_y=0;
void mouse(int button,int state,int x,int y) 
{ /* mouse being pressed or released--save position for motion */
	mouse_x=x; mouse_y=y;
}
void motion(int x, int y) { 
	float scale=0.01; /* radians rotation/frame per mouse pixel motion */
	orient.nudge(
		-scale*(x-mouse_x), 
		scale*(y-mouse_y)
	);
	mouse_x=x; mouse_y=y; /* save old mouse positions */
}
void keyboard(unsigned char key,int x,int y) {
	key_down[key]=true;
}
void keyboardUp(unsigned char key,int x,int y) {
	key_down[key]=false;
}

int main(int argc,char **argv) {
	glutInit(&argc,argv);  /* set up GLUT */
	glutInitWindowPosition(0,60);
	glutInitWindowSize(1000,700);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH);
	glutCreateWindow("Particle Demo (try WASD keys!)"); /* make a window */
	glewInit(); /* only AFTER create window! */
	glutDisplayFunc(display); /* here's how to draw the window */
	glutMouseFunc(mouse);
	glutMotionFunc(motion); 
	glutKeyboardFunc(keyboard);
	glutKeyboardUpFunc(keyboardUp);
	glutMainLoop(); /* start drawing */
}
